package at.jku.ssw.set.tree.test;

import java.util.ArrayList;
import java.util.Arrays;

import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import at.jku.ssw.set.impl.InOrderTreeIterator;
import at.jku.students.set.tree.BinaryTreeSet;

public class BinaryTreeSetTest {

	protected BinaryTreeSet createTreeSet(char... objects) {
		BinaryTreeSet set = new BinaryTreeSet();
		for (char c : objects) {
			set.add(c);
		}
		return set;
	}

	private static ArrayList<Character> subList(ArrayList<Character> oldList, int startIndex, int lastIndex) {
		ArrayList<Character> newList = new ArrayList<>();
		for (int i = startIndex; i < lastIndex; i++) {
			newList.add(oldList.get(i));
		}
		return newList;
	}

	public static ArrayList<ArrayList<Character>> permutation(ArrayList<Character> s) {
		ArrayList<ArrayList<Character>> res = new ArrayList<ArrayList<Character>>();
		if (s.size() == 1) {
			res.add(s);
		} else if (s.size() > 1) {
			int lastIndex = s.size() - 1;
			ArrayList<Character> last = subList(s, lastIndex, s.size());
			ArrayList<Character> rest = subList(s, 0, lastIndex);
			res = merge(permutation(rest), last);
		}
		return res;
	}

	public static ArrayList<ArrayList<Character>> merge(ArrayList<ArrayList<Character>> list,
			ArrayList<Character> last) {
		ArrayList<ArrayList<Character>> res = new ArrayList<>();
		for (ArrayList<Character> s : list) {
			for (int i = 0; i <= s.size(); ++i) {
				ArrayList<Character> ps = new ArrayList<>(s);
				for (int j = 0; j < last.size(); j++) {
					ps.add(i + j, last.get(j));
				}
				res.add(ps);
			}
		}
		return res;
	}

	@Rule
	public ExpectedException thrown = ExpectedException.none();

	@Test
	public void testAdd() {
		ArrayList<Character> originalList = new ArrayList<>(Arrays.asList('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'j'));
		for (ArrayList<Character> l : permutation(originalList)) {
			BinaryTreeSet set = new BinaryTreeSet();
			for (Character i : l) {
				set.add(i);
			}
			InOrderTreeIterator it = new InOrderTreeIterator(set.getRoot());
			ArrayList<Character> integers = new ArrayList<>();
			while (it.hasNext()) {
				integers.add((Character) it.next());
			}
			Assert.assertEquals(originalList, integers);
		}
	}

	@Test
	public void testAddRemove0() {
		ArrayList<Character> originalList = new ArrayList<>(Arrays.asList('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i'));
		ArrayList<Character> toRemove = new ArrayList<>(Arrays.asList('a', 'b', 'c', 'd'));
		ArrayList<Character> toRemoveMissing = new ArrayList<>(Arrays.asList('#'));
		ArrayList<Character> removedList = new ArrayList<>(Arrays.asList('e', 'f', 'g', 'h', 'i'));
		for (ArrayList<Character> l : permutation(originalList)) {
			BinaryTreeSet set = new BinaryTreeSet();
			for (Character i : l) {
				set.add(i);
			}
			for (Character r : toRemove) {
				Assert.assertTrue(set.remove(r));
			}
			for (Character r : toRemoveMissing) {
				Assert.assertFalse(set.remove(r));
			}
			InOrderTreeIterator it = new InOrderTreeIterator(set.getRoot());
			ArrayList<Character> characters = new ArrayList<>();
			while (it.hasNext()) {
				characters.add((Character) it.next());
			}
			Assert.assertEquals(removedList, characters);
		}
	}

	@Test
	public void testAddRemove1() {
		char[] insertions = { 'd', 'a', 'b' };
		BinaryTreeSet set = new BinaryTreeSet();
		for (char value : insertions) {
			set.add(value);
		}
		Assert.assertTrue(set.remove('b'));
	}

	@Test
	public void testEmptyHeight() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.height());
	}

	@Test
	public void testHeight1() {
		BinaryTreeSet set = createTreeSet('a', 'b', 'c', 'd', 'e', 'f');
		Assert.assertEquals(6, set.height());
	}

	@Test
	public void testHeight2() {
		BinaryTreeSet set = createTreeSet('a', 'b', 'c', 'd', 'e', ' ');
		Assert.assertEquals(5, set.height());
	}

	@Test
	public void testEmptyWdith() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.width());
	}

	@Test
	public void testWidth() {
		BinaryTreeSet set = createTreeSet('d', 'b', 'a', 'c', 'f', 'e', 'g', (char) 0);
		Assert.assertEquals(4, set.width());
	}

	@Test
	public void testSize() {
		BinaryTreeSet set = createTreeSet('d', 'b', 'a', 'c', 'f', 'e', 'g', (char) 0);
		Assert.assertEquals(8, set.size());
	}

	@Test
	public void testSizeEmpty() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.size());
	}

	@Test
	public void testAddContains() {
		BinaryTreeSet set = createTreeSet((char) 0);
		set.add((char) 0);
		Assert.assertEquals(1, set.size());
		Assert.assertEquals(true, set.contains((char) 0));
	}

	@Test
	public void testTreeString() {
		BinaryTreeSet set = new BinaryTreeSet();
		set.add('b');
		set.add('c');
		set.add('a');
		String s = set.treeString();

		StringBuilder sb = new StringBuilder();
		sb.append("=> b").append(System.lineSeparator());
		sb.append("  => a").append(System.lineSeparator());
		sb.append("  => c").append(System.lineSeparator());
		Assert.assertEquals(sb.toString(), s);

	}

}
